 /////////////////////////////////////////////////////////////////////////
 //                 LLIBRERIES DEL COORDINADOR                          //
/////////////////////////////////////////////////////////////////////////

#include <RF24Network.h>
#include <RF24.h>
#include <SPI.h>
  /////////////////////////////////////////////////////////////////////////
 //                          PARAMETRES I VARIABLES                     //
/////////////////////////////////////////////////////////////////////////

// Pins SPI declaració del NRF24
RF24 radio(7, 8);
RF24Network network(radio);

// Adreça del node actual en octal
const uint16_t aquest_node = 00;

// Adreça node a enviar en octal
const uint16_t altre_node = 01;

// Variables del payload a rebre
struct dades_t {
  float Acc;
  float BatS;
  float SolS;
  float Id;
};

// Variables a adquirir en el coordinador
float BatC;
float SolC;

// Matrius de caràcters per a la concatenació
char charsols[10];
char charbats[10];
char characc[10];
char charbatc[10];
char charsolc[10];
char charid[10];

// Entrada analògica de la placa solar
int sensorsolar = A1;

// Declaració tupla per guardar el payload a rebre
dades_t dades;

// Comptadors
int t4;

  ///////////////////////////////////////////////////////////////////////// 
 //       CONFIGURACIÓ INICIAL DEL COORDINADOR                          // 
/////////////////////////////////////////////////////////////////////////
void setup(void) {
  
  // Inicialització port série
  Serial.begin(115200);

  // Pin digital per a despertar el ESP8266 
  pinMode(3, OUTPUT);
 
  // Inicialització NRF24
  SPI.begin();
  radio.begin();
  network.begin(90, aquest_node);



  // Velocitat de transmissió radio del NRF24
  radio.setDataRate(RF24_250KBPS);
  
}

  ///////////////////////////////////////////////////////////////////////// 
 //           BUCLE PRINCIPAL COORDINADOR                               // 
/////////////////////////////////////////////////////////////////////////

void loop(void) {
    
  // Reset variables  
  dades.Id=0.0; 
  
  // Comptador  
  t4 = t4 + 1;

  // Si s’ha arribat a la repetició 5
  if (t4 > 4) {

    // Despertar NRF24 
    radio.powerUp();
    radio.begin();
    network.begin(90, altre_node);
    network.update();                          
    SPI.begin();

         // Escoltar si arriba algun paquet
    while ( network.available() ) {     
       RF24NetworkHeader header;        
       bool ok = network.read(header, &dades, sizeof(dades));

       // Si dades.Id>1.0 ha arribat un nou paquet
       if (dades.Id > 1.0) {
          
      // Despertar el ESP8266 activant el pin connectat a CH_PD
          digitalWrite(3, HIGH);
      delayMicroseconds(100);

          // Inici port serie per enviar les dades al ESP8266
          Serial.begin(115200);

          // El paquet no es d’accident
      if (dades.Acc < 10.0) {

              //Crida la funció llegirVcc() llegir voltatge intern
              BatC = 1126400 / (llegirVcc() * 10);
              BatC = round(BatC);
              BatC = BatC / 100;

              //Lectura pin analogic del voltatge de la placa solar
              SolC  = analogRead(sensorsolar);
              SolC = SolC * 4.0 / 1023;
          
          // El paquet es d’accident
          } else if (dades.Acc > 10.0) {

              //Assignem 0.0 a les variables SolC i BatC
      BatC = 0.0;
        SolC = 0.0;
          }

          // Espais i lletres a concatenar
          String stringbats = "";
          String stringsols = "";
          String stringbatc = "";
          String stringsolc = "";
    String stringacc = "";
    String stringid = "";
          String bats = "B";
          String sols = "S";
          String batc = "C";
          String solc = "T";
      String acc = "A";
          String idd = "I";

          // Concatenació variable dades.BatS i el prefix B
          dtostrf(dades.BatS, 5, 2, charbats);
          for (int i = 0; i < sizeof(charbats); i++)
          {
              stringbats += charbats[i];
          }
          bats += stringbats;
          bats.remove(5, 10);

          // Impressió del string bats pel port serie 
          Serial.println(bats);
    
          // Concatenació i impressio dades.SolS i el prefix S
          dtostrf(dades.SolS, 5, 2, charsols);
          for (int i = 0; i < sizeof(charsols); i++)
          {
              stringsols += charsols[i];
          }
          sols += stringsols;
          sols.remove(5, 10);
          Serial.println(sols);
 
          // Concatenació i impressió BatC i el prefix C
          dtostrf(BatC, 5, 2, charbatc);
          for (int i = 0; i < sizeof(charbatc); i++)
          {
              stringbatc += charbatc[i];
          }
          batc += stringbatc;
          batc.remove(5, 10);
          Serial.println(batc);

          // Concatenació i impressió SolC i el prefix T 
          dtostrf(SolC, 5, 2, charsolc);
          for (int i = 0; i < sizeof(charsolc); i++)
          {
              stringsolc += charsolc[i];
          }
          solc += stringsolc;
          solc.remove(5, 10);
          Serial.println(solc);

                // Concatenació i impressió dades.Acc i el prefix A
        dtostrf(dades.Acc, 5, 2, characc);
          for (int i = 0; i < sizeof(characc); i++)
          {
              stringacc += characc[i];
          }
          acc += stringacc;
          acc.remove(5, 10);
          Serial.println(acc);
   
      // Concatenació dades.Id i el prefix I
          dtostrf(dades.Id, 5, 2, charid);
          for (int i = 0; i < sizeof(charid); i++)
          {
              stringid += charid[i];
          }
          idd += stringid;
          idd.remove(5, 10);
          Serial.println(idd);
  
      // Desalimentar el pin en mode power down
           digitalWrite(3, LOW);
           delayMicroseconds(100);
     }
    }

    // Adormir el NRF24 i apagar el port serie
    SPI.end();
    radio.powerDown();
    Serial.end();

    // Resetejar comptador t4
    t4=0;
  }
}

  ///////////////////////////////////////////////////////////////////////// 
 //            FUNCIONS DEL COORDINADOR                                 // 
/////////////////////////////////////////////////////////////////////////

// Lectura interna del voltatge d’alimentació
long llegirVcc() {
  long result; // Read 1.1V reference against AVcc
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1); delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA, ADSC)); result = ADCL; result |= ADCH << 8; result = 1126400L / result; // Back-calculate AVcc in mV return result;
}

